@@ -7,6 +7,10 @@ from typing import Iterable
 from typing import Optional
 from typing import Set
 from typing import Tuple
+from typing import Sequence
+import itertools
+
+import networkx as nx  # type: ignore
 
 _port_id_counter = 0
 _vertex_id_counter = 0
@@ -105,10 +109,10 @@ class Vertex(object):
     Every vertex representes a main element in a ForSyDe model.
     Every vertex contains a number of ports (which are repeated in the
     vertexed to increase reliability in the model, since putting
-    them in edges would have been sufficient) with their associated types.
-    Also, every port contains "Properties" which are arbitrary associated data,
-    such as the size of bits in a Signal or the time slots in a Time Division
-    Multiplexer.
+    them in edges would have been sufficient).
+    Also, every vertex contains "Properties" which are arbitrary slef-contained 
+    associated data, such as the size of bits in a Signal or the time slots in 
+    a Time Division Multiplexer.
     """
 
     identifier: str = field(default_factory=_generate_vertex_id, hash=True)
@@ -137,21 +141,41 @@ class Vertex(object):
             raise AttributeError(
                 f"Required port {name} of {self.identifier} does not exist.")
 
-    def get_neigh(self, name: str, model) -> "Vertex":
-        out_port = self.get_port(name)
-        for n in model.adj[self]:
-            for (_, edata) in model.edges[self][n]:
+    def get_neigh(self, name: str, model, out_dir=True, in_dir=False) -> Optional["Vertex"]:
+        port = self.get_port(name)
+        neighs = iter(())
+        if out_dir:
+            neighs = itertools.chain(neighs, (n for n in model[self]))
+        if in_dir:
+            neighs = itertools.chain(neighs, (n for n in model if self in model[n]))
+        for n in neighs:
+            for (_, edata) in model.get_edge_data(self, n).items():
                 edge = edata["object"]
-                if edge.source_vertex_port == out_port:
+                if edge.source_vertex_port == port or n.target_vertex_port == port:
                     return edge.target_vertex
+        return None
+
+    def get_neighs(self, name: str, model) -> Sequence["Vertex"]:
+        port = self.get_port(name)
+        neighs = iter()
+        if out_dir:
+            neighs = itertools.chain(neighs, (n for n in model[self]))
+        if in_dir:
+            neighs = itertools.chain(neighs, (n for n in model if self in model[n]))
+        for n in neighs:
+            for (_, edata) in model.get_edge_data(self, n).items():
+                edge = edata["object"]
+                if edge.source_vertex_port == port or n.target_vertex_port == port:
+                    yield edge.target_vertex
+
 
 
 @dataclass
 class Edge(object):
     """Class containing all information for an Edge.
 
-    The edge contains references to the source and target vetexes
-    as well as the ports being connect on both ends, in case
+    The edge contains references to the source and target 'Vertex'es
+    as well as the 'Port's being connect on both ends, in case
     they exist. The edges also have types associated with them
     so that extra deductions can be made along the EDA flow.
     """
@@ -178,3 +202,41 @@ class Edge(object):
     # def is_type(self, tsource: ModelType, ttarget: ModelType) -> bool:
     #     return self.source_vertex.is_type(
     #         tsource) and self.target_vertex.is_type(ttarget)
+
+
+class ForSyDeModel(nx.MultiDiGraph):
+    """The main graph holder element representing a ForSyDe Model
+
+    A subclass of MultiDiGraph from the networkX library, this class
+    holds the model (a graph model therefore) which can be used for
+    any purpose in the ForSyDe design flow. In addition to all standard
+    graph algorithms and facilities given by networkX, this subclass
+    also provides parsing and dumping from the canonical ForSyDe IO
+    disk persistent file and other output formats that can be used for
+    visualization, such as GraphML.
+
+    It also provides additional methods building on top of standard graph
+    methods to make development easier, such as directly iterating vertexes
+    by their associated types.
+    """
+    def __init__(self,
+                 standard_views=[
+                     'create_tables.sql', 'types.sql', 'create_views.sql'
+                 ],
+                 *args,
+                 **kwargs):
+        nx.MultiDiGraph.__init__(self, *args, **kwargs)
+
+    def neighs(self, v: Vertex) -> Iterable[Vertex]:
+        yield from self.nodes.adj[v]
+
+    def neighs_rev(self, v: Vertex) -> Iterable[Vertex]:
+        yield from nx.reverse_view(self).adj[v]
+
+    def get_vertex(self,
+                   label: str,
+                   label_name: str = 'label') -> Optional[Vertex]:
+        for (v, d) in self.nodes.data():
+            if d[label_name] == label:
+                return v
+        return None
