@@ -1,7 +1,16 @@
+import abc
+import warnings
+import xml.etree.ElementTree as etree
+import pprint
+import itertools
+from functools import reduce
 from typing import Any
 from typing import Optional
 
-from lxml import etree  # type: ignore
+# from lxml import etree  # type: ignore
+import networkx as nx
+import pydot
+#from networkx.drawing.nx_pydot import write_dot  # type: ignore
 
 from forsyde.io.python.core import Vertex
 from forsyde.io.python.core import Edge
@@ -11,10 +20,12 @@ from forsyde.io.python.types import VertexFactory
 from forsyde.io.python.types import EdgeFactory
 
 
-class ForSyDeModelDriver:
+class ForSyDeModelDriver(abc.ABC):
+    @abc.abstractmethod
     def write(self, model: ForSyDeModel, sink: str) -> None:
         pass
 
+    @abc.abstractmethod
     def read(
             self,
             source: str,
@@ -63,8 +74,8 @@ class ForSyDeMLDriver(ForSyDeModelDriver):
 
     def write(self, model: ForSyDeModel, sink: str) -> None:
         xmlns = '{http://graphml.graphdrawing.org/xmlns}'
-        tree = etree.Element(xmlns + 'graphml')
-        graph = etree.SubElement(tree, xmlns + 'graph')
+        root = etree.Element(xmlns + 'graphml')
+        graph = etree.SubElement(root, xmlns + 'graph')
         graph.set('id', 'model')
         graph.set('edgedefault', 'directed')
         for v in model.nodes:
@@ -103,31 +114,29 @@ class ForSyDeMLDriver(ForSyDeModelDriver):
             if edge.target_vertex_port:
                 edge_elem.set('targetport', edge.target_vertex_port.identifier)
         with open(sink, 'wb') as sinkstream:
-            sinkstream.write(
-                etree.tostring(tree,
-                               pretty_print=True,
-                               xml_declaration=True,
-                               encoding='UTF-8'))
+            tree = etree.ElementTree(element=root)
+            tree.write(sinkstream, encoding="utf-8", xml_declaration=True)
+            #sinkstream.write(etree.tostring(tree, pretty_print=True, xml_declaration=True, encoding='UTF-8'))
 
     def read(self,
              source: str,
              other_model: Optional[ForSyDeModel] = None) -> ForSyDeModel:
+        xmlns = '{http://graphml.graphdrawing.org/xmlns}'
         model = ForSyDeModel() if not other_model else other_model
         with open(source, 'r') as instream:
             tree = etree.parse(instream)
-            for vnode in tree.xpath('/xmlns:graphml//xmlns:graph/xmlns:node',
-                                    namespaces=self.ns):
+            for vnode in tree.findall('.//' + xmlns + 'graph/' + xmlns +
+                                      'node'):
                 vertex = VertexFactory.build(identifier=vnode.get('id'),
                                              type_name=vnode.get('type'))
                 model.add_node(vertex, label=vertex.identifier)
-                for portnode in vnode.xpath("xmlns:port", namespaces=self.ns):
+                for portnode in vnode.iterfind('./' + xmlns + "port"):
                     port = Port(identifier=portnode.get('name'))
                     vertex.ports.add(port)
                 dataopen = [(vnode, vertex.properties)]
                 while len(dataopen) > 0:
                     (parent, data) = dataopen.pop()
-                    for datanode in parent.xpath('xmlns:data',
-                                                 namespaces=self.ns):
+                    for datanode in parent.iterfind(xmlns + 'data'):
                         dataname = datanode.get('attr.name')
                         datatype = self.str_to_type(datanode.get('attr.type'))
                         if datatype is str\
@@ -144,14 +153,23 @@ class ForSyDeMLDriver(ForSyDeModelDriver):
                             else:
                                 data[dataname] = child
                             dataopen.append((datanode, child))
-            for vedge in tree.xpath('/xmlns:graphml//xmlns:graph/xmlns:edge',
-                                    namespaces=self.ns):
-                source_vertex = next(n
-                                     for (n, nid) in model.nodes.data('label')
-                                     if nid == vedge.get('source'))
-                target_vertex = next(n
-                                     for (n, nid) in model.nodes.data('label')
-                                     if nid == vedge.get('target'))
+            for vedge in tree.iterfind('.//' + xmlns + 'graph/' + xmlns +
+                                       'edge',
+                                       namespaces=self.ns):
+                source_vertex = next((n
+                                      for (n, nid) in model.nodes.data('label')
+                                      if nid == vedge.get('source')), None)
+                if source_vertex is None:
+                    raise ValueError(
+                        f"Could not find source node {vedge.get('source')} to during edge building."
+                    )
+                target_vertex = next((n
+                                      for (n, nid) in model.nodes.data('label')
+                                      if nid == vedge.get('target')), None)
+                if target_vertex is None:
+                    raise ValueError(
+                        f"Could not find target node {vedge.get('target')} to during edge building."
+                    )
                 edge = EdgeFactory.build(source=source_vertex,
                                          target=target_vertex,
                                          type_name=vedge.get('type'))
@@ -171,6 +189,11 @@ class ForSyDeMLDriver(ForSyDeModelDriver):
 
 
 class ForSyDeXMLDriver(ForSyDeModelDriver):
+    def __init__(self) -> None:
+        warnings.warn(
+            "The 'ForSyDeXMLDriver' for 'xml' python driver is deprecated. Use 'ForSyDeMLDriver' for 'forxml' instead.",
+            DeprecationWarning)
+
     def property_to_xml(self, parent: etree.Element, prop: Any) -> None:
         '''Transform an object into the expected XML element layout
         
@@ -205,23 +228,24 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
         It collects the children recursively of 'elem' to the expected
         formats and types.
         '''
-        if len(elem.xpath("Property[@index]")) > 0:
-            elems = elem.xpath("Property[@index]")
+        if len(elem.iterfind("Property[@index]")) > 0:
+            elems = elem.iterfind("Property[@index]")
             l = [None for i in range(len(elems))]
             for e in elems:
                 l[int(e.get('index'))] = self.xml_to_property(e)
             return l
-        elif len(elem.xpath("Property[@name]")) > 0:
+        elif len(elem.iterfind("Property[@name]")) > 0:
             return {
                 e.get("name"): self.xml_to_property(e)
-                for e in elem.xpath("Property[@name]")
+                for e in elem.iterfind("Property[@name]")
             }
-        elif len(elem.xpath("Property")) > 0:
-            return set(self.xml_to_property(e) for e in elem.xpath("Property"))
+        elif len(elem.iterfind("Property")) > 0:
+            return set(
+                self.xml_to_property(e) for e in elem.iterfind("Property"))
         elif elem.text and elem.text.strip():
-            if elem.xpath('Property[@type="Integer"]'):
+            if elem.iterfind('Property[@type="Integer"]'):
                 return int(elem.text.strip())
-            elif elem.xpath('Property[@type="Float"]'):
+            elif elem.iterfind('Property[@type="Float"]'):
                 return float(elem.text.strip())
             else:
                 return elem.text.strip()
@@ -229,6 +253,8 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
             return dict()
 
     def write(self, model: ForSyDeModel, sink: str) -> None:
+        raise NotImplementedError(
+            "Writing to ad-hoc XML format is not supported.")
         tree = etree.Element('ForSyDe')
         for v in model.nodes:
             node_elem = etree.SubElement(tree, 'Vertex')
@@ -237,14 +263,14 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
             for port in v.ports:
                 port_elem = etree.SubElement(node_elem, 'Port')
                 port_elem.set('id', port.identifier)
-                if port.port_type == int:
-                    port_elem.set('type', 'Integer')
-                elif port.port_type == float:
-                    port_elem.set('type', 'Integer')
-                elif issubclass(port.port_type, Vertex):
-                    port_elem.set('type', str(port.port_type))
-                else:
-                    port_elem.set('type', 'String')
+                # if port.port_type == int:
+                #     port_elem.set('type', 'Integer')
+                # elif port.port_type == float:
+                #     port_elem.set('type', 'Integer')
+                # elif issubclass(port.port_type, Vertex):
+                #     port_elem.set('type', str(port.port_type))
+                # else:
+                #     port_elem.set('type', 'String')
             for (prop, val) in v.properties.items():
                 prop_elem = etree.SubElement(node_elem, 'Property')
                 prop_elem.set('name', prop)
@@ -271,11 +297,11 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
         model = ForSyDeModel() if not other_model else other_model
         with open(source, 'r') as instream:
             tree = etree.parse(instream)
-            for vnode in tree.xpath('/ForSyDeModel/Vertex'):
+            for vnode in tree.findall('./Vertex'):
                 vertex = VertexFactory.build(identifier=vnode.get('id'),
                                              type_name=vnode.get('type'))
                 model.add_node(vertex, label=vertex.identifier)
-                for portnode in vnode.xpath("Port"):
+                for portnode in vnode.iterfind("Port"):
                     port = Port(identifier=portnode.get('id'))
                     if portnode.get('type') == "Integer":
                         port.port_type = int
@@ -289,12 +315,12 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
                             port.port_type = str
                     vertex.ports.add(port)
                 vertex.properties = self.xml_to_property(vnode)
-                # for propnode in vnode.xpath('Property'):
+                # for propnode in vnode.iterfind('Property'):
                 #     prop_name = propnode.get('name')
                 #     prop_val = model.xml_to_property(
                 #         propnode)  # json.loads(propnode.text)
                 #     vertex.properties[prop_name] = prop_val
-            for vedge in tree.xpath('/ForSyDeModel/Edge'):
+            for vedge in tree.findall('./Edge'):
                 source_vertex = next(n
                                      for (n, nid) in model.nodes.data('label')
                                      if nid == vedge.get('source_id'))
@@ -319,3 +345,60 @@ class ForSyDeXMLDriver(ForSyDeModelDriver):
                                key=key,
                                object=edge)
         return model
+
+
+class ForSyDeDotDriver(ForSyDeModelDriver):
+    '''Dot writer for some visualization of the models.
+    '''
+    def read(self, source: str,
+             other_model: Optional[ForSyDeModel]) -> Optional[ForSyDeModel]:
+        raise NotImplementedError(
+            "The 'DotDriver' only supports writing the model to 'dot'.")
+
+    def write(self, model: ForSyDeModel, sink: str) -> None:
+        dot = pydot.Dot("model", graph_type="digraph")
+        # create all clusters
+        graphs = {'': dot}
+        for v in model.nodes:
+            graph_name = "/".join(v.identifier.split("/")[:-1])
+            label = f'{v.identifier}\n'
+            label += v.get_type_tag() + "\\n"
+            label += pprint.pformat(v.properties)
+            graph = graphs[graph_name] if graph_name in graphs else pydot.Subgraph(
+                    f"cluster_{graph_name}", label=graph_name, color="black")
+            graph.add_node(
+                pydot.Node(v.identifier, label=label, shape='box'))
+            while graph_name not in graphs:
+                graphs[graph_name] = graph
+                graph_name = "/".join(graph_name.split("/")[:-1])
+                parent = graphs[graph_name] if graph_name in graphs else pydot.Subgraph(f"cluster_{graph_name}",
+                                                     label=graph_name,
+                                                     color="black")
+                parent.add_subgraph(graph)
+                graph = parent
+        for (s, t, e) in model.edges.data("object"):
+            sp = e.source_vertex_port
+            tp = e.target_vertex_port
+            graph_name_s = "/".join(s.identifier.split("/")[:-1])
+            graph_name_t = "/".join(t.identifier.split("/")[:-1])
+            g = dot
+            if graph_name_s == graph_name_t:
+                g = graphs[graph_name_s]
+            g.add_edge(
+                pydot.Edge(s.identifier,
+                           t.identifier,
+                           label=f"{e.get_type_tag()}\\n" +
+                           (f"{s.identifier}.{sp.identifier}"
+                            if sp else f"{s.identifier}") + "\\n" +
+                           (f"{t.identifier}.{tp.identifier}"
+                            if tp else f"{t.identifier}")))
+        # for ((g1_name, g1), (g2_name, g2)) in itertools.product(graphs.items(), graphs.items()):
+        #     if g1 != g2 and g2_name.startswith(g1_name):
+        #         print(g1_name, g2_name)
+        #         g2.subgraph(g1)
+        # for (_, g) in graphs.items():
+        #     if g != dot:
+        #         dot.subgraph(g)
+        with open(sink, "w") as sinkstream:
+            sinkstream.write(dot.to_string())
+        # write_dot(strg, sink)
